@@ -84,6 +84,15 @@ static const char *dwarf_tag_names[] = {
 	[DW_TAG_type_unit]		  = "type_unit",
 	[DW_TAG_rvalue_reference_type]    = "rvalue_reference_type",
 #endif
+#if _ELFUTILS_PREREQ(0, 170)
+	[DW_TAG_coarray_type]		  = "coarray_type",
+	[DW_TAG_generic_subrange]	  = "generic_subrange",
+	[DW_TAG_dynamic_type]		  = "dynamic_type",
+	[DW_TAG_call_site]		  = "call_site",
+	[DW_TAG_call_site_parameter]	  = "call_site_parameter",
+	[DW_TAG_skeleton_unit]		  = "skeleton_unit",
+	[DW_TAG_immutable_type]		  = "immutable_type",
+#endif
 };
 
 static const char *dwarf_gnu_tag_names[] = {
@@ -107,10 +116,14 @@ static const char *dwarf_gnu_tag_names[] = {
 const char *dwarf_tag_name(const uint32_t tag)
 {
 	if (tag >= DW_TAG_array_type && tag <=
+#if _ELFUTILS_PREREQ(0, 170)
+		DW_TAG_immutable_type
+#else
 #ifdef STB_GNU_UNIQUE
 		DW_TAG_rvalue_reference_type
 #else
 		DW_TAG_shared_type
+#endif
 #endif
 	    )
 		return dwarf_tag_names[tag];
@@ -127,7 +140,7 @@ const char *dwarf_tag_name(const uint32_t tag)
 	return "INVALID";
 }
 
-static const struct conf_fprintf conf_fprintf__defaults = {
+static struct conf_fprintf conf_fprintf__defaults = {
 	.name_spacing = 23,
 	.type_spacing = 26,
 	.emit_stats   = 1,
@@ -135,11 +148,10 @@ static const struct conf_fprintf conf_fprintf__defaults = {
 
 const char tabs[] = "\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t";
 
-static size_t cacheline_size;
 
-size_t tag__nr_cachelines(const struct tag *tag, const struct cu *cu)
+size_t tag__nr_cachelines(const struct conf_fprintf *conf, const struct tag *tag, const struct cu *cu)
 {
-	return (tag__size(tag, cu) + cacheline_size - 1) / cacheline_size;
+	return (tag__size(tag, cu) + conf->cacheline_size - 1) / conf->cacheline_size;
 }
 
 static const char *tag__accessibility(const struct tag *tag)
@@ -611,7 +623,7 @@ static size_t type__fprintf_stats(struct type *type, const struct cu *cu,
 {
 	size_t printed = fprintf(fp, "\n%.*s/* size: %d, cachelines: %zd, members: %u",
 				 conf->indent, tabs, type->size,
-				 tag__nr_cachelines(type__tag(type), cu), type->nr_members);
+				 tag__nr_cachelines(conf, type__tag(type), cu), type->nr_members);
 
 	if (type->nr_static_members != 0)
 		printed += fprintf(fp, ", static members: %u */\n", type->nr_static_members);
@@ -787,7 +799,7 @@ print_default:
 					   (type->tag == DW_TAG_class_type &&
 					    !tconf.classes_as_structs) ? "class" : "struct",
 					   tconf.type_spacing - 7,
-					   type__name(ctype), name);
+					   type__name(ctype), name ?: "");
 		} else {
 			struct class *cclass = tag__class(type);
 
@@ -802,7 +814,7 @@ print_default:
 		ctype = tag__type(type);
 
 		if (type__name(ctype) != NULL && !expand_types) {
-			printed += fprintf(fp, "union %-*s %s", tconf.type_spacing - 6, type__name(ctype), name);
+			printed += fprintf(fp, "union %-*s %s", tconf.type_spacing - 6, type__name(ctype), name ?: "");
 		} else {
 			tconf.type_spacing -= 8;
 			printed += union__fprintf(ctype, cu, &tconf, fp);
@@ -812,7 +824,7 @@ print_default:
 		ctype = tag__type(type);
 
 		if (type__name(ctype) != NULL)
-			printed += fprintf(fp, "enum %-*s %s", tconf.type_spacing - 5, type__name(ctype), name);
+			printed += fprintf(fp, "enum %-*s %s", tconf.type_spacing - 5, type__name(ctype), name ?: "");
 		else
 			printed += enumeration__fprintf(type, &tconf, fp);
 		break;
@@ -837,6 +849,7 @@ static size_t class_member__fprintf(struct class_member *member, bool union_memb
 				     struct conf_fprintf *conf, FILE *fp)
 {
 	const int size = member->byte_size;
+	int member_alignment_printed = 0;
 	struct conf_fprintf sconf = *conf;
 	uint32_t offset = member->byte_offset;
 	size_t printed = 0, printed_cacheline = 0;
@@ -863,7 +876,21 @@ static size_t class_member__fprintf(struct class_member *member, bool union_memb
 	if (member->is_static)
 		printed += fprintf(fp, "static ");
 
-	printed += type__fprintf(type, cu, name, &sconf, fp);
+	/* For struct-like constructs, the name of the member cannot be
+	 * conflated with the name of its type, otherwise __attribute__ are
+	 * printed in the wrong order.
+	 */
+	if (tag__is_union(type) || tag__is_struct(type) ||
+	    tag__is_enumeration(type)) {
+		printed += type__fprintf(type, cu, NULL, &sconf, fp);
+		if (name) {
+			if (!type__name(tag__type(type)))
+				printed += fprintf(fp, " ");
+			printed += fprintf(fp, "%s", name);
+		}
+	} else {
+		printed += type__fprintf(type, cu, name, &sconf, fp);
+	}
 
 	if (member->is_static) {
 		if (member->const_value != 0)
@@ -872,8 +899,10 @@ static size_t class_member__fprintf(struct class_member *member, bool union_memb
 		printed += fprintf(fp, ":%u", member->bitfield_size);
 	}
 
-	if (!sconf.suppress_aligned_attribute && member->alignment != 0)
-		printed += fprintf(fp, " __attribute__((__aligned__(%u)))", member->alignment);
+	if (!sconf.suppress_aligned_attribute && member->alignment != 0) {
+		member_alignment_printed = fprintf(fp, " __attribute__((__aligned__(%u)))", member->alignment);
+		printed += member_alignment_printed;
+	}
 
 	fputc(';', fp);
 	++printed;
@@ -884,7 +913,7 @@ static size_t class_member__fprintf(struct class_member *member, bool union_memb
 	    type__name(tag__type(type)) == NULL) {
 		if (!sconf.suppress_offset_comment) {
 			/* Check if this is a anonymous union */
-			int slen = cm_name ? (int)strlen(cm_name) : -1;
+			int slen = member_alignment_printed + (cm_name ? (int)strlen(cm_name) : -1);
 			int size_spacing = 5;
 
 			if (tag__is_struct(type) && tag__class(type)->is_packed && !conf->suppress_packed) {
@@ -892,12 +921,6 @@ static size_t class_member__fprintf(struct class_member *member, bool union_memb
 				slen += packed_len;
 			}
 
-			if (tag__type(type)->alignment != 0 && !conf->suppress_aligned_attribute) {
-				char bftmp[64];
-				int aligned_len = snprintf(bftmp, sizeof(bftmp), " __attribute__((__aligned__(%u)))", tag__type(type)->alignment);
-				slen += aligned_len;
-			}
-
 			printed += fprintf(fp, sconf.hex_fmt ?
 							"%*s/* %#5x" :
 							"%*s/* %5u",
@@ -1293,11 +1316,11 @@ static size_t class__fprintf_cacheline_boundary(struct conf_fprintf *conf,
 						FILE *fp)
 {
 	int indent = conf->indent;
-	uint32_t cacheline = offset / cacheline_size;
+	uint32_t cacheline = offset / conf->cacheline_size;
 	size_t printed = 0;
 
 	if (cacheline > *conf->cachelinep) {
-		const uint32_t cacheline_pos = offset % cacheline_size;
+		const uint32_t cacheline_pos = offset % conf->cacheline_size;
 		const uint32_t cacheline_in_bytes = offset - cacheline_pos;
 
 		if (cacheline_pos == 0)
@@ -1739,7 +1762,7 @@ static size_t __class__fprintf(struct class *class, const struct cu *cu,
 		}
 		printed += fprintf(fp, " */\n");
 	}
-	cacheline = (cconf.base_offset + type->size) % cacheline_size;
+	cacheline = (cconf.base_offset + type->size) % conf->cacheline_size;
 	if (cacheline != 0)
 		printed += fprintf(fp, "%.*s/* last cacheline: %u bytes */\n",
 				   cconf.indent, tabs,
@@ -1966,15 +1989,22 @@ static long cacheline__size(void)
 #endif
 }
 
-void dwarves__fprintf_init(uint16_t user_cacheline_size)
+void dwarves__resolve_cacheline_size(const struct conf_load *conf, uint16_t user_cacheline_size)
 {
+	uint16_t size;
+
 	if (user_cacheline_size == 0) {
 		long sys_cacheline_size = cacheline__size();
 
 		if (sys_cacheline_size > 0)
-			cacheline_size = sys_cacheline_size;
+			size = sys_cacheline_size;
 		else
-			cacheline_size = 64; /* Fall back to a sane value */
+			size = 64; /* Fall back to a sane value */
 	} else
-		cacheline_size = user_cacheline_size;
+		size = user_cacheline_size;
+
+	if (conf)
+		conf->conf_fprintf->cacheline_size = size;
+
+	conf_fprintf__defaults.cacheline_size = size;
 }
